from torch.utils.data import Dataset
import pandas as pd


class Q2Dataset(Dataset):
    def __init__(self, file):
        """
        读取数据文件，并封装成dataset
        Args:
            file: 数据文件
        """
        super(Q2Dataset, self).__init__()
        df = pd.read_csv(file, sep='\t', header=0, na_filter=False)
        self.samples = df['q2'].map(lambda x: ''.join(x.split()))

    def __getitem__(self, item):
        return self.samples[item]

    def __len__(self):
        return len(self.samples)



# 测试线上服务
import os
import json
import time
import argparse
import requests
from flask import Flask, request
app = Flask(__name__)

import turbo_transformers
import torch
from model.turbo_match import ErnieMatchTurbo
from recall.semantic_search import FaissStreamSearcher
from recall.es_search import ElasticObj
from model.ernie import convert_list_to_features, new_tensor


keywords_url = 'http://10.18.217.184:8077/queryprocv2?userId=yuewenhao&type=TV&query='
# 从主服务获取tag和keyword labeling作为关键词
def get_keywords(sentence):
    kw = []
    try:
        temp = requests.get(keywords_url + sentence)
        body = temp.json()
        tag_result = body['SemanticResult']['tagResults']
        for tw in tag_result:
            kw.append(tw['word'])
        labeling = body['KeywordLabeling']['keyword']
        kw.extend(labeling)
        kw = list(set(kw))
    except:
        return []
    return kw


# 输入单个数据预测
@app.route('/chat', methods=['GET'])
def get_final():
    context = request.args.get('context')
    us = context.strip().split(sep)
    kw_q2 = get_keywords(us[-1])
    # 召回
    _, topk, t_vec, t_pca, t_search = indexer.recall(us[-1:], pca, index, k_index)
    topk = list(map(str, topk[0]))
    t0 = time.time()
    if not kw_q2:
        responses1, origins1 = es.recall(us[-1], topk, k_es)
    else:
        responses1, origins1 = es.recall(' '.join(kw_q2), topk, k_es)
    t01 = time.time()
    responses2, origins2 = es.recall_ids(topk[:10])
    responses = responses1 + responses2
    origins = origins1 + origins2
    t1 = time.time()
    # rank
    if not responses:
        best, score = 'NA', 0
        t11 = t1
        t12 = t1
    else:
        query = '。'.join(us)
        input_ids, input_mask, segment_ids = convert_list_to_features(query, 30, turbo_model.tokenizer, responses)
        t11 = time.time()
        input_ids = new_tensor(input_ids)
        input_mask = new_tensor(input_mask)
        segment_ids = new_tensor(segment_ids)
        logits = turbo_model(input_ids, input_mask, segment_ids)
        logits = torch.sigmoid(logits)
        score = torch.max(logits).item()
        pred = torch.argmax(logits).item()
        best = responses[pred]
    t2 = time.time()
    return json.dumps({'response': best,
                       'score': score,
                       'keywords': kw_q2,
                       'candidates': [o + sep + r for o, r in zip(origins, responses)],
                       'time': t2 - t0 + t_vec + t_pca + t_search,
                       't_vec': t_vec,
                       't_pca': t_pca,
                       't_search': t_search,
                       't_esid': t1 - t01,
                       't_esrecall': t01 - t0,
                       't_input': t11 - t1,
                       't_model': t2 - t11,
                       't_rank': t2-t1
                       })


# test the whole system
if __name__ == '__main__':
    config_file = "../config/config.json"
    with open(config_file, 'r') as f:
        config_info = json.load(f)
    k_index = config_info['recall']['index_topn']
    k_es = config_info['recall']['es_topn']
    sep = "||"

    os.environ["CUDA_VISIBLE_DEVICES"] = '7'

    parser = argparse.ArgumentParser()
    # 注意local rank这个参数，必须要以这种形式指定，即使代码中不使用。因为 launch 工具默认传递该参数
    parser.add_argument("--local_rank", type=int, default=0)
    parser.add_argument("--num_gpu", type=int, default=4)
    args = parser.parse_args()

    es = ElasticObj(config_info)

    turbo_transformers.set_num_threads(4)
    turbo_model = ErnieMatchTurbo(config_info)

    indexer = FaissStreamSearcher(args, config_info)
    tt0 = time.time()
    index, pca = indexer.build_ivf()
    tt1 = time.time()
    print(f'build index take {tt1-tt0} seconds')

    app.run(host='0.0.0.0', port=6996, threaded=True)



import torch
from transformers import BertTokenizer, BertModel
import sys

sys.path.append("..")
import numpy as np


class InputExample(object):
    """
    构造的输入对象，包括id和内容。
    Attributes:
        unique_id: 编号。
        a: 第一句
        b: 第二句
    """

    def __init__(self, unique_id, a, b=None):
        self.unique_id = unique_id
        self.a = a
        self.b = b


class InputFeatures(object):
    """
    构造的特征对象。
    Attributes:
        input_ids: 输入的token。
        segment_ids: 用于区分不同句子的序号。
        input_mask: mask。
    """

    def __init__(self, input_ids, segment_ids, input_mask):
        self.input_ids = input_ids
        self.segment_ids = segment_ids
        self.input_mask = input_mask


def new_tensor(array, requires_grad=False):
    """
    构建一个新的不参与反向传播的张量，如果gpu可用就放置到gpu上。
    Args:
        array: 待构建向量的原始数据。
        requires_grad: 是否需要参与反向传播。
    Returns:
        构建好的张量。
    """
    if torch.cuda.is_available():
        tensor = torch.as_tensor(array, device=torch.cuda.current_device())
        tensor.requires_grad_(requires_grad)
    else:
        tensor = torch.as_tensor(array)
        tensor.requires_grad_(requires_grad)
    return tensor


def read_examples(a_list, b_list=None):
    """
    读取句子构造InputExample对象。
    Args:
        a_list: 样本第一句列表。
        b_list: 样本第二句列表。
    Returns:
        构造的InputExample对象集。
    """
    examples = []
    unique_id = 0
    if b_list:
        for a, b in zip(a_list, b_list):
            examples.append(InputExample(unique_id, a, b))
            unique_id += 1
        return examples
    else:
        for a in a_list:
            examples.append(InputExample(unique_id, a))
            unique_id += 1
        return examples


def convert_examples_to_features(examples, max_seq_length, tokenizer):
    """
    把一系列InputExample对象转换为一系列InputFeatures对象。
    Args:
        examples: InputExample对象集。
        max_seq_length: 最大长度。
        tokenizer: 分词函数。
    Returns:
        转换得到的InputFeatures对象集，以及所有语句序列的token。
    """
    all_features = []
    all_tokens = []
    for example in examples:
        feature, tokens = convert_single_example(example, max_seq_length, tokenizer)
        all_features.append(feature)
        all_tokens.append(tokens)
    return all_features, all_tokens


def convert_list_to_features(a, max_seq_length, tokenizer, b_list):
    """
    把句子列表转换为特征id。适用于匹配阶段，更快。
    Args:
        a: 第一句。
        max_seq_length: 最大长度。
        tokenizer: 分词函数。
        b_list: 第二句列表。
    Returns:
        转换得到的所有样本的输入id，mask id，以及segment id。
    """
    a_tokens = tokenizer.tokenize(a)
    a_len = len(a_tokens)
    b_tokens_list = list(map(tokenizer.tokenize, b_list))
    max_half = max_seq_length // 2
    input_ids = []
    input_masks = []
    segment_ids = []
    for b_tokens in b_tokens_list:
        b_len = len(b_tokens)
        if a_len > b_len:
            a_len_trunc = max(max_half, max_seq_length - b_len)
            b_len_trunc = min(max_half, b_len)
        else:
            a_len_trunc = min(max_half, a_len)
            b_len_trunc = max(max_half, max_seq_length - a_len)
        # 截断前面
        a_tokens_trunc = a_tokens[-a_len_trunc:]
        # 截断后面
        b_tokens_trunc = b_tokens[:b_len_trunc]
        tokens = ["[CLS]"] + a_tokens_trunc + ["[SEP]"] + b_tokens_trunc + ["[SEP]"]
        ids = tokenizer.convert_tokens_to_ids(tokens)
        ids_len = len(ids)
        pad_ids = ids + [0] * (max_seq_length + 3 - ids_len)
        input_ids.append(pad_ids)
        mask = [1] * ids_len + [0] * (max_seq_length + 3 - ids_len)
        input_masks.append(mask)
        seg = [0] * (len(a_tokens_trunc) + 2) + [1] * (len(b_tokens_trunc) + 1) + [0] * (max_seq_length + 3 - ids_len)
        segment_ids.append(seg)

    input_ids = np.stack(input_ids, axis=0)
    input_masks = np.stack(input_masks, axis=0)
    segment_ids = np.stack(segment_ids, axis=0)

    return input_ids, input_masks, segment_ids


def _truncate_seq_pair(tokens_a, tokens_b, max_length):
    total_length = len(tokens_a) + len(tokens_b)
    while True:
        if total_length <= max_length:
            break
        if len(tokens_a) > len(tokens_b):
            tokens_a.pop()
        else:
            tokens_b.pop()
        total_length -= 1


def convert_single_example(example, max_seq_length, tokenizer):
    """
    把一个InputExample对象转换为一个InputFeatures对象。
    Args:
        example: 一个InputExample对象。
        max_seq_length: 最大长度。
        tokenizer: 分词函数。
    Returns:
        处理后的Inputfeature对象，和对应序列的id表示。
    """
    a_tokens = tokenizer.tokenize(example.a)
    b_tokens = None
    if example.b:
        b_tokens = tokenizer.tokenize(example.b)
    if b_tokens:
        _truncate_seq_pair(a_tokens, b_tokens, max_seq_length)
    else:
        a_tokens = a_tokens[:max_seq_length]

    tokens = ["[CLS]"]
    segment_ids = [0]
    for token in a_tokens:
        tokens.append(token)
        segment_ids.append(0)
    tokens.append("[SEP]")
    segment_ids.append(0)

    if b_tokens:
        for token in b_tokens:
            tokens.append(token)
            segment_ids.append(1)
        tokens.append("[SEP]")
        segment_ids.append(1)

    input_ids = tokenizer.convert_tokens_to_ids(tokens)

    # The mask has 1 for real tokens and 0 for padding tokens. Only real
    # tokens are attended to.
    input_mask = [1] * len(input_ids)

    # Zero-pad up to the sequence length.
    if b_tokens:
        pad_length = 3
    else:
        pad_length = 2
    while len(input_ids) < max_seq_length + pad_length:
        input_ids.append(0)
        input_mask.append(0)
        segment_ids.append(0)

    assert len(input_ids) == max_seq_length + pad_length
    assert len(input_mask) == max_seq_length + pad_length
    assert len(segment_ids) == max_seq_length + pad_length

    feature = InputFeatures(
        input_ids=input_ids,
        input_mask=input_mask,
        segment_ids=segment_ids
    )
    return feature, tokens


def infer_batch(a_batch, model, tokenizer, max_seq_len, b_batch=None):
    """
    用一个批次的数据对模型进行训练。
    Args:
        a_batch: 一个批次的样本第一句(required)。
        b_batch: 一个批次的样本第二句(optional)。
        model: 模型(required)。
        tokenizer: 模型(required)。
        max_seq_len: context长度限制(required)。
    Returns:
        预测每个样本。
    """
    # batch_size个InputExample对象
    predict_examples = read_examples(a_batch, b_batch)
    # 转换为batch_size个feature对象
    features, all_tokens = convert_examples_to_features(predict_examples, max_seq_len, tokenizer)

    # input_ids [bs_size, seq_len]
    input_ids = new_tensor([f.input_ids for f in features])
    # input_mask [bs_size, seq_len]
    input_mask = new_tensor([f.input_mask for f in features])
    # segment_ids [bs_size, seq_len]
    segment_ids = new_tensor([f.segment_ids for f in features])
    output = model(input_ids, mask=input_mask, seg=segment_ids)
    return output


class ErnieVectorizer(torch.nn.Module):
    """
    不带head的ernie预训练模型，提取特征向量
    """

    def __init__(self):
        super(ErnieVectorizer, self).__init__()
        # 加载预训练ernie模型
        self.tokenizer = BertTokenizer.from_pretrained("nghuyong/ernie-1.0")
        self.pretrain = BertModel.from_pretrained("nghuyong/ernie-1.0")

    def forward(self, x, mask=None, seg=None):
        sequence_output, pooled_output = self.pretrain(x, attention_mask=mask, token_type_ids=seg)
        return torch.mean(sequence_output, 1)


class ErnieMatch(torch.nn.Module):
    """
    带有全连接head的ernie模型用于微调匹配任务
    """

    def __init__(self):
        super(ErnieMatch, self).__init__()
        # 加载预训练ernie模型
        self.tokenizer = BertTokenizer.from_pretrained("nghuyong/ernie-1.0")
        self.pretrain = BertModel.from_pretrained("nghuyong/ernie-1.0")
        self.dropout = torch.nn.Dropout(0.1)
        self.fc1 = torch.nn.Linear(768, 512)
        self.fc2 = torch.nn.Linear(512, 1)

    def forward(self, x, mask=None, seg=None):
        sequence_output, pooled_output = self.pretrain(x, attention_mask=mask, token_type_ids=seg)
        output = self.dropout(pooled_output)
        output = self.fc1(output)
        output = self.fc2(output).squeeze()
        return output



# 语义检索召回
import numpy as np
import requests
import json
from sklearn.decomposition import IncrementalPCA
import pickle
from tqdm import tqdm
import time
import os
import sys
import pandas as pd

sys.path.append("..")
import torch
from torch.utils.data.distributed import DistributedSampler
import argparse
import torch.backends.cudnn as cudnn
import faiss
from model.ernie import infer_batch, ErnieVectorizer
from data_loader.v1_data import Q2Dataset
from data_loader.v2_data import Q2StreamSubset


class FaissSearcher:
    def __init__(self, args, config):
        super().__init__()
        self.num_gpu = args.num_gpu
        self.batch_size = config['recall']['batch_size']
        self.max_seq_len = config['recall']['max_seq_len']
        # 候选文件
        self.candidate_file = config['inputs']['candidate_file']
        # 加载编码模型
        self.model = ErnieVectorizer()
        self.tokenizer = self.model.tokenizer
        # put model to GPU if available
        self.local_rank = args.local_rank
        torch.cuda.set_device(self.local_rank)
        if torch.cuda.is_available():
            self.model = self.model.cuda()
            self.model = torch.nn.parallel.DistributedDataParallel(self.model,
                                                                   device_ids=[self.local_rank],
                                                                   output_device=self.local_rank,
                                                                   find_unused_parameters=True)

    def candidate_vectorize(self):
        """
        调用编码模型批量预测所有候选句子的特征向量，可分布式
        :return: 返回候选句子总数。向量保存为npy文件。
        """
        # 分布式训练，全量数据读取
        dataset = Q2Dataset(self.candidate_file)
        vecs = []
        with torch.no_grad():
            if torch.cuda.is_available():
                print('using pytorch DistributedSampler')
                sampler = DistributedSampler(dataset, shuffle=False)  # 只能兼容map style dataset
                loader = torch.utils.data.DataLoader(dataset, batch_size=self.batch_size, sampler=sampler,
                                                     pin_memory=True)
            else:
                loader = torch.utils.data.DataLoader(dataset, batch_size=self.batch_size, shuffle=False,
                                                     pin_memory=True)
            for data in loader:
                vec = infer_batch(data,
                                  self.model,
                                  self.tokenizer,
                                  self.max_seq_len)
                if torch.cuda.is_available():
                    vec = vec.cpu()
                vecs.append(vec)
        vecs = torch.cat(vecs).numpy()
        if torch.cuda.is_available():
            np.save(f'candidates_part{self.local_rank}', vecs)
        else:
            np.save('candidates', vecs)

        assert len(dataset) <= vecs.shape[0] * self.num_gpu < len(dataset) + self.num_gpu, \
            f'num of vectors is {vecs.shape[0]}, but num of sentences is {len(dataset)}'
        return len(dataset)

    def vec_gen(self):
        """
        单进程vector生成器，从多个npy文件中轮流读取vector并流式生成。本身不控制生成结束。
        :return: yield一个vector
        """
        if torch.cuda.is_available():
            cycle = 0  # 轮数，用来同步遍历每个部分文件
            vec_all = []
            for i in range(self.num_gpu):
                vec_all.append(np.load(f'candidates_part{i}.npy', mmap_mode='r'))
            while 1:
                for i in range(self.num_gpu):
                    yield vec_all[i][cycle, :]
                cycle += 1
        else:
            vec_all = np.load('candidates.npy', mmap_mode='r')
            for i in range(vec_all.shape[0]):
                yield vec_all[i, :]

    def build_index(self, n_candidates):
        """
        构建faiss索引，仅单进程
        :param n_candidates: 候选总数
        :return: 索引
        """
        rows = []
        vecs = self.vec_gen()
        for i in range(n_candidates):
            rows.append(next(vecs))
        arr = np.stack(rows, axis=0).astype('float32')
        res = faiss.StandardGpuResources()  # use a single GPU
        dim = 768
        assert arr.shape == (n_candidates, dim)
        measure = faiss.METRIC_L2
        param = 'PCA128, IVF10000, PQ16'
        vec_index = faiss.index_factory(dim, param, measure)
        if torch.cuda.is_available():
            gpu_index = faiss.index_cpu_to_gpu(res, 0, vec_index)
            print(gpu_index.is_trained)
            gpu_index.train(arr)
            gpu_index.add(arr)
            print('total rows: ', gpu_index.ntotal)
            return gpu_index
        else:
            print(vec_index.is_trained)
            vec_index.train(arr)
            vec_index.add(arr)
            print('total rows: ', vec_index.ntotal)
            return vec_index

    def recall(self, query, index):
        """
        在faiss索引中检索
        :param query: 一批原始查询语句（不分词，仅q2）
        :param index: 索引
        :return:
        """
        index.nprobe = 10
        with torch.no_grad():
            q_vecs = infer_batch(query, self.model, self.tokenizer, self.max_seq_len)
            if torch.cuda.is_available():
                q_vecs = q_vecs.cpu()
            q_vecs = q_vecs.numpy()
        k = 10  # topk
        distance, topk = index.search(q_vecs, k)
        dataset = Q2Dataset(self.candidate_file)
        for i, q in enumerate(query):
            print('query: ', q)
            for j, idx in enumerate(topk[i]):
                print(f'{j}th nearest candidate: ', dataset[idx])


class FaissStreamSearcher:
    def __init__(self, num_gpu, local_rank, config):
        """
        先用模型对输入语句编码成向量，再用faiss进行向量检索。
        使用流式读写，用IncrementalPCA降维，增量构建faiss索引，适用于大规模数据。
        :param :
        """
        super().__init__()
        self.num_gpu = num_gpu
        self.batch_size = config['recall']['batch_size']
        self.max_seq_len = config['recall']['max_seq_len']
        # 候选文件
        self.candidate_file = config['inputs']['candidate_file']
        # PCA模型
        self.pca_dim = config['recall']['pca_dim']
        self.pca_checkpoint = config['recall']['pca_checkpoint']
        # 加载编码模型
        self.model = ErnieVectorizer()
        self.tokenizer = self.model.tokenizer
        # 在线编码地址 todo
        self.vectorizer_url = config['recall']['vectorizer_url']
        self.candidate_vecs = config['recall']['candidate_vecs']
        # put model to GPU if available
        self.local_rank = local_rank
        # torch.cuda.set_device(self.local_rank)
        if torch.cuda.is_available():
            self.model = self.model.cuda()

    def candidate_vectorize(self):
        """
        调用编码模型批量预测所有候选句子的特征向量，可分布式
        :return: 返回候选句子总数。向量保存为vec文件。
        """
        # 分布式训练，流式数据读取
        dataset = Q2StreamSubset(self.candidate_file, self.local_rank, self.num_gpu)
        with torch.no_grad():
            loader = torch.utils.data.DataLoader(dataset, batch_size=self.batch_size, shuffle=False, pin_memory=True)
            if torch.cuda.is_available():
                # 平均每个文件接近10G
                file = f'candidates_part{self.local_rank}.vec'
            else:
                file = 'candidates.vec'

            with open(file, 'ab') as fw:
                for data in loader:
                    vec = infer_batch(data,
                                      self.model,
                                      self.tokenizer,
                                      self.max_seq_len)
                    if torch.cuda.is_available():
                        vec = vec.cpu()
                    pickle.dump(vec.numpy(), fw)

    def _interleave_sort(self, batches):
        """
        交错排序，恢复原始顺序，多个批次整合为一个批次
        :param batches: 若干批的 numpy array
        :return: 恢复顺序之后的统一批次的 numpy array
        """
        arr = []
        col = len(batches)
        for i in range(self.batch_size):
            for j in range(self.num_gpu):
                if j < col and i < batches[j].shape[0]:
                    arr.append(batches[j][i, :])
                else:
                    return np.stack(arr, axis=0).astype('float32')
        return np.stack(arr, axis=0).astype('float32')

    def batch_vec_gen(self):
        """
        单进程batch vector生成器
        从多个向量文件中轮流读取vector，按原始顺序拼成批vectors并流式生成。本身可以控制生成结束。
        批的数量：batch size * num gpu（最后一批除外）
        :return: yield 批vectors
        """
        if torch.cuda.is_available():
            file_handlers = []
            for i in range(self.num_gpu):
                # todo 配置
                file_handlers.append(open(f'/disk2/yangshansong/candidates_part{i}.vec', 'rb'))
            while True:
                batch_vecs = []
                try:
                    # 每个文件读取一个批，一轮共num gpu个批
                    for i in range(self.num_gpu):
                        # dump阶段序列化多个批对象，load阶段每次反序列化一个批对象
                        sub = pickle.load(file_handlers[i])
                        print('sub length: ', len(sub))
                        batch_vecs.append(sub)
                    print('batch_vecs length: ', len(batch_vecs))
                    # 生成器，批数据的数量是：batch_size * num_gpu
                    yield self._interleave_sort(batch_vecs)
                except EOFError:
                    # 流终止：最后一批可能不是每个文件都有，可能每个文件的shape不一样
                    if batch_vecs:
                        yield self._interleave_sort(batch_vecs)
                    break
        else:
            with open('candidates.vec', 'rb') as fr:
                while True:
                    try:
                        yield pickle.load(fr)
                    except EOFError:
                        break

    def pca_fit(self):
        """
        增量训练PCA模型，并保存
        Returns:
        """
        vec_gen = self.batch_vec_gen()
        pca = IncrementalPCA(n_components=self.pca_dim)
        # todo
        count = 1
        for batch_vec in tqdm(vec_gen):
            if count % 100 == 0:
                print('count: ', count)
            pca.partial_fit(batch_vec)
            count += 1
            if count == 1000:
                break
        with open(self.pca_checkpoint, 'wb') as fw:
            pickle.dump(pca, fw)
        print('PCA model trained and saved!')

    def pca_transform(self):
        """
        使用训练好的增量PCA模型对所有向量降维，并保存为一个npy文件
        Returns:
        """
        # 加载增量PCA模型
        with open(self.pca_checkpoint, 'rb') as fr:
            pca = pickle.load(fr)

        vec_gen = self.batch_vec_gen()
        vecs = []
        for batch_vec in tqdm(vec_gen):
            # PCA降维
            batch_vec = pca.transform(batch_vec)
            print('batch_vec: ', batch_vec)
            vecs.append(batch_vec)
            print('----------------')

        vecs = np.concatenate(vecs, axis=0).astype('float32')
        np.save('candidates_small.npy', vecs)
        print('vectors reduced and saved!')

    def build_ivf(self):
        """
        构建ivf索引，仅单进程。
        一次性加入向量，先用PCA模型降维，再加入索引。
        :return: 索引， pca模型
        """
        # 加载增量PCA模型
        with open(self.pca_checkpoint, 'rb') as fr:
            pca_model = pickle.load(fr)
            print("has loaded pca model")
        # 构建索引
        res = faiss.StandardGpuResources()  # use a single GPU
        measure = faiss.METRIC_L2
        param = 'IVF10000, PQ16'
        vec_index = faiss.index_factory(self.pca_dim, param, measure)
        if torch.cuda.is_available():
            vec_index = faiss.index_cpu_to_gpu(res, 0, vec_index)
        # 训练并加入向量
        vecs = np.load(self.candidate_vecs)
        vec_index.train(vecs)
        vec_index.add(vecs)
        print('index total rows: ', vec_index.ntotal)
        return vec_index, pca_model

    def build_hnsw(self):
        """
        构建hnsw索引，仅单进程。
        流式加入向量，先用PCA模型降维，再加入索引。
        :return: 索引，pca模型
        """
        # 加载增量PCA模型
        with open(self.pca_checkpoint, 'rb') as fr:
            pca_model = pickle.load(fr)
        # 构建索引
        res = faiss.StandardGpuResources()  # use a single GPU
        measure = faiss.METRIC_L2
        param = 'HNSW16'
        vec_index = faiss.index_factory(self.pca_dim, param, measure)
        if torch.cuda.is_available():
            vec_index = faiss.index_cpu_to_gpu(res, 0, vec_index)
        # 加入向量
        vecs = np.load(self.candidate_vecs)
        vec_index.add(vecs)
        print('index total rows: ', vec_index.ntotal)
        return vec_index, pca_model

    def recall(self, query, pca_model, index, k):
        """
        向量检索，Faiss语义索引
        :param query: 一批原始查询语句（不分词，仅q2）
        :param pca_model: 训练好的pca模型
        :param index: 索引
        :param k: 召回数量
        :return:
        """
        with torch.no_grad():
            t0 = time.time()
            # 查询向量化，通过ERNIE模型，获得查询的向量
            query_vecs = infer_batch(query, self.model, self.tokenizer, self.max_seq_len)
            if torch.cuda.is_available():
                query_vecs = query_vecs.cpu()
            query_vecs = query_vecs.numpy()
        t1 = time.time()
        # PCA降维
        query_vecs = pca_model.transform(query_vecs).astype('float32')
        t2 = time.time()
        # 语义检索
        distance, top = index.search(query_vecs, k)
        t3 = time.time()
        return distance, top, t1 - t0, t2 - t1, t3 - t2

    def recall_online(self, query, pca_model, index, k):
        """
        Faiss索引中检索，使用文本向量化服务
        :param query: 一个原始查询语句（不分词，仅q2）
        :param pca_model: 训练好的pca模型
        :param index: 索引
        :param k: 召回数量
        :return:
        """
        t0 = time.time()
        # 通过接口，获取查询向量化
        temp = requests.get(self.vectorizer_url + query)
        q_vecs = temp.json()
        q_vecs = np.reshape(q_vecs, (1, -1))
        t1 = time.time()
        q_vecs = pca_model.transform(q_vecs).astype('float32')
        t2 = time.time()
        distance, top = index.search(q_vecs, k)
        t3 = time.time()
        return distance, top, t1 - t0, t2 - t1, t3 - t2


def search_test():
    """
    基于向量化后的聊天数据库，通过Faiss，建立语义近邻索引，并搜索
    Returns:
    """
    indexer = FaissStreamSearcher(args, config_info)
    t0 = time.time()
    index, pca_model = indexer.build_hnsw()
    t1 = time.time()
    print(f'build index take {t1 - t0} seconds')
    qs = [
        '券商股疯了，大盘也是疯了，比疫情期间大跌还害怕。',
        '你好啊',
        '很高兴认识你',
        '我要去听相声了德云社的哦',
        '今年利物浦确实有冠军的实力'
    ]
    dist, tops, _, _, _ = indexer.recall(qs, pca_model, index, 1000)
    for dis in dist:
        row = pd.Series(dis)
        print(row.describe())


# python -m torch.distributed.launch --nproc_per_node=4 semantic_search.py
if __name__ == "__main__":
    os.environ["CUDA_VISIBLE_DEVICES"] = '4,5,6,7'
    parser = argparse.ArgumentParser()
    # 注意local rank这个参数，必须要以这种形式指定，即使代码中不使用。因为 launch 工具默认传递该参数
    parser.add_argument("--local_rank", type=int)
    parser.add_argument("--num_gpu", type=int, default=4)
    args = parser.parse_args()
    local_rank = args.local_rank
    num_gpu = args.num_gpu

    config_file = "../config/config.json"
    with open(config_file, 'r') as f:
        config_info = json.load(f)

    # 如果gpu可用
    if torch.cuda.is_available():
        # 初始化分布式进程组
        torch.distributed.init_process_group(backend='NCCL', init_method='env://')
    # 令cuDNN可以使用非确定性算法
    cudnn.enabled = True
    # 令cuDNN的auto-tuner自动寻找最适合当前配置的高效算法
    cudnn.benchmark = True
    # 使用确定性算法
    cudnn.deterministic = True

    indexer = FaissStreamSearcher(local_rank, num_gpu, config_info)
    # indexer.candidate_vectorize()
    # indexer.pca_fit()
    # indexer.pca_transform()
    search_test()



import numpy as np
import tensorrt as trt
import pycuda.driver as cuda
import pycuda.autoinit


class ONNXClassifierWrapper:
    def __init__(self, file, out_size, target_dtype=np.float32):
        self.target_dtype = target_dtype
        self.out_size = out_size
        self.load(file)
        self.stream = None

    def load(self, file):
        f = open(file, "rb")
        runtime = trt.Runtime(trt.Logger(trt.Logger.WARNING))
        engine = runtime.deserialize_cuda_engine(f.read())
        self.context = engine.create_execution_context()

    def allocate_memory(self, input1, input2, input3):
        # Need to set both input and output precisions to FP16 to fully enable FP16
        self.output = np.empty(self.out_size, dtype=self.target_dtype)

        # Allocate device memory
        self.d_input1 = cuda.mem_alloc(1 * input1.nbytes)
        self.d_input2 = cuda.mem_alloc(1 * input2.nbytes)
        self.d_input3 = cuda.mem_alloc(1 * input3.nbytes)
        self.d_output = cuda.mem_alloc(1 * self.output.nbytes)

        self.bindings = [int(self.d_input1), int(self.d_input2), int(self.d_input3), int(self.d_output)]

        self.stream = cuda.Stream()

    def predict(self, input1, input2, input3):  # result gets copied into output
        if self.stream is None:
            self.allocate_memory(input1, input2, input3)

        # Transfer input data to device
        cuda.memcpy_htod_async(self.d_input1, input1, self.stream)
        cuda.memcpy_htod_async(self.d_input2, input2, self.stream)
        cuda.memcpy_htod_async(self.d_input3, input3, self.stream)
        # Execute model
        self.context.execute_async_v2(self.bindings, self.stream.handle, None)
        # Transfer predictions back
        cuda.memcpy_dtoh_async(self.output, self.d_output, self.stream)
        # Syncronize threads
        self.stream.synchronize()

        return self.output


if __name__ == '__main__':
    import os
    import json
    import sys
    sys.path.append('../')
    from transformers import BertTokenizer
    from rank.ranker import ErnieRanker
    from model.ernie import convert_list_to_features, new_tensor

    os.environ["CUDA_VISIBLE_DEVICES"] = '5'

    batch_size = 20
    trt_model = ONNXClassifierWrapper("ernie_match.trt", batch_size)

    # config_file = "../config/config.json"
    # with open(config_file, 'r') as f:
    #     config_info = json.load(f)
    # torch_model = ErnieRanker(config_info)
    # torch_model.load_model()

    # predict after loading the model
    query = '请问'
    response = ['哈哈', '你好', '什么东西？', '呵呵哒', '你要问啥？', '哈哈', '你好', '什么东西？', '呵呵哒', '你要问啥？', '哈哈', '你好', '什么东西？', '呵呵哒', '你要问啥？', '哈哈', '你好', '什么东西？', '呵呵哒', '你要问啥？']
    tokenizer = BertTokenizer.from_pretrained("nghuyong/ernie-1.0")
    input_ids, input_mask, segment_ids = convert_list_to_features(query, 30, tokenizer, response)
    input_ids = np.asarray(input_ids, dtype=np.int32)
    input_mask = np.asarray(input_mask, dtype=np.int32)
    segment_ids = np.asarray(segment_ids, dtype=np.int32)

    trt_result = trt_model.predict(input_ids, input_mask, segment_ids)
    print(trt_result)

    # input_ids = new_tensor(input_ids)
    # input_mask = new_tensor(input_mask)
    # segment_ids = new_tensor(segment_ids)
    #
    # # compare
    # torch_model.model.eval()
    # torch_result = torch_model.model(input_ids, input_mask, segment_ids)
    # print(torch_result.detach().cpu().numpy())



# 转化模型为onnx格式
import os
import numpy as np
import torch
import onnx
import onnxruntime
from onnxruntime.transformers import optimizer
from onnxruntime.transformers.onnx_model_bert import BertOptimizationOptions
from model.ernie import ErnieMatch, read_examples, convert_examples_to_features, new_tensor

os.environ["CUDA_VISIBLE_DEVICES"] = '4'

# 加载模型
pkl_checkpoint = "../resources/17.pkl"
torch_model = ErnieMatch()
if torch.cuda.is_available():
    torch_model = torch_model.cuda()

if os.path.exists(pkl_checkpoint):
    torch_model.load_state_dict(torch.load(pkl_checkpoint, map_location='cpu'))
else:
    raise FileNotFoundError(f'the resources file {pkl_checkpoint} is not found!')

# set the model to inference mode
torch_model.eval()

# 任意输入，固定维度
max_seq_len = 30
batch_size = 20
context = ['你好'] * batch_size
response = ['再见'] * batch_size
predict_examples = read_examples(context, response)
features, all_tokens = convert_examples_to_features(predict_examples, max_seq_len, torch_model.tokenizer)
# input_ids [bs_size, seq_len]
input_ids = new_tensor([f.input_ids for f in features])
# input_mask [bs_size, seq_len]
input_mask = new_tensor([f.input_mask for f in features])
# segment_ids [bs_size, seq_len]
segment_ids = new_tensor([f.segment_ids for f in features])
torch_out = torch_model(input_ids, mask=input_mask, seg=segment_ids)

# 进行转化
torch.onnx.export(torch_model,  # model being run
                  (input_ids, input_mask, segment_ids),  # model input (or a tuple for multiple inputs)
                  "ernie_match.onnx",  # where to save the model (can be a file or file-like object)
                  export_params=True,  # store the trained parameter weights inside the model file
                  opset_version=11,  # the ONNX version to export the model to
                  do_constant_folding=True,  # whether to execute constant folding for optimization
                  keep_initializers_as_inputs=False,
                  input_names=['input_ids', 'input_mask', 'segment_ids'],  # the model's input names
                  output_names=['output'])  # the model's output names

# optimized_model = optimizer.optimize_model("ernie_match_raw.onnx", model_type='bert', num_heads=12, hidden_size=768, opt_level=99, use_gpu=True)
# # optimized_model.convert_model_float32_to_float16()  # only for gpu with tensor core
# optimized_model.save_model_to_file("ernie_match.onnx")


onnx_model = onnx.load("ernie_match.onnx")
onnx.checker.check_model(onnx_model)

so = onnxruntime.SessionOptions()
so.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_ALL
ort_session = onnxruntime.InferenceSession("ernie_match.onnx", sess_options=so)
# ort_session = onnxruntime.InferenceSession("ernie_match.onnx")
ort_session.set_providers(['CUDAExecutionProvider'])
ort_inputs = {
    ort_session.get_inputs()[0].name: input_ids.cpu().numpy(),
    ort_session.get_inputs()[1].name: input_mask.cpu().numpy(),
    ort_session.get_inputs()[2].name: segment_ids.cpu().numpy()
}
ort_outs = ort_session.run(None, ort_inputs)

# compare ONNX Runtime and PyTorch results
np.testing.assert_allclose(torch_out.detach().cpu().numpy(), ort_outs[0], rtol=1e-03, atol=1e-05)
print("Exported model has been tested with ONNXRuntime, and the result looks good!")



import os
import torch
from model.ernie import ErnieMatch, infer_batch


class ErnieRanker:
    def __init__(self, config):
        """
        精排模型
        :param args:
        """
        super().__init__()
        self.max_seq_len = config['rank']['max_seq_len']
        self.pkl = config['rank']['match_checkpoint']
        # 加载模型
        self.model = ErnieMatch()
        self.tokenizer = self.model.tokenizer
        # put model to GPU if available
        if torch.cuda.is_available():
            self.model = self.model.cuda()

    def load_model(self):
        """
        加载一个模型文件
        :return: 无返回值。
        """
        if os.path.exists(self.pkl):
            self.model.load_state_dict(torch.load(self.pkl, map_location='cpu'))
        else:
            raise FileNotFoundError(f'the checkpoint file {self.pkl} is not found!')
        self.model.eval()

    def predict(self, contexts, responses):
        """
        调用模型批量预测所有历史和回复对的match分数，选择最高的一个
        :param contexts: 一批历史
        :param responses: 一批回复
        :return: 最好的一个回复。
        """
        logits = infer_batch(contexts, self.model, self.tokenizer, self.max_seq_len, responses)
        logits = torch.sigmoid(logits)
        score = torch.max(logits).item()
        pred = torch.argmax(logits).item()
        return responses[pred], score


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
"""
@File        :  __init__.py.py
@Project     :  ptmchat_develop
@Contact     :  yangshansong
@License     :  (C)Copyright 2021-2022, Hisense
@Description :  None

@Modify Time        @Author         @Version        @Description
------------        -------         --------        ------------
2021/12/6 11:24     yangshansong    1.0             None
"""



import turbo_transformers
import json
import os
import torch
from model.ernie import ErnieMatch


class ErnieMatchTurbo:
    """
    带有全连接head的ernie模型用于微调匹配任务
    turbo transformer加速推理版，直接加载微调好的整个模型的参数
    bert_model是turbo类，其他层是torch类
    """
    def __init__(self, config):
        self.pkl = config['rank']['match_checkpoint']
        self.model = ErnieMatch()
        self.tokenizer = self.model.tokenizer
        # put model to GPU if available
        if torch.cuda.is_available():
            self.model = self.model.cuda()
        # 加载参数
        if os.path.exists(self.pkl):
            self.model.load_state_dict(torch.load(self.pkl, map_location='cpu'))
        else:
            raise FileNotFoundError(f'the checkpoint file {self.pkl} is not found!')
        self.model.eval()

        self.bert_model = turbo_transformers.BertModel.from_torch(self.model.pretrain)
        self.dropout = self.model.dropout
        self.fc1 = self.model.fc1
        self.fc2 = self.model.fc2

    def __call__(self, x, mask=None, seg=None):
        sequence_output, pooled_output = self.bert_model(x, attention_masks=mask, token_type_ids=seg)
        output = self.dropout(pooled_output)
        output = self.fc1(output)
        output = self.fc2(output).squeeze()
        return output


if __name__ == '__main__':
    from rank.ranker import ErnieRanker
    from model.ernie import convert_list_to_features, new_tensor
    os.environ["CUDA_VISIBLE_DEVICES"] = '2'
    # use 4 threads for BERT inference
    turbo_transformers.set_num_threads(4)

    # the initialization of the acceleration model
    config_file = "../config/config.json"
    with open(config_file, 'r') as f:
        config_info = json.load(f)
    turbo_model = ErnieMatchTurbo(config_info)

    # predict after loading the model
    context = '你好'
    response = ['再见'] * 20
    input_ids, input_mask, segment_ids = convert_list_to_features(context, 30, turbo_model.tokenizer, response)
    input_ids = new_tensor(input_ids)
    input_mask = new_tensor(input_mask)
    segment_ids = new_tensor(segment_ids)

    turbo_result = turbo_model(input_ids, input_mask, segment_ids)

    # compare
    torch_model = ErnieRanker(config_info)
    torch_model.load_model()
    torch_result = torch_model.model(input_ids, input_mask, segment_ids)
    print(turbo_result)
    print(torch_result)



# 测试线上服务
import os
import json
import time
import argparse
import requests
from flask import Flask, request
import sys

app = Flask(__name__)
sys.path.append("/tmp/pycharm_project_201/ptmchat_develop/")
from recall.semantic_search import FaissStreamSearcher
from recall.es_search import ElasticObj
from rank.ranker import ErnieRanker

es = None
sep = "||"
index = None
pca = None
k_index = None
k_es = None
indexer = None
matcher = None

keywords_url = 'http://10.18.217.184:8077/queryprocv2?userId=yuewenhao&type=TV&query='


def get_keywords(sentence):
    """
    获取语句的关键词，依赖于语义主服务
    Args:
        sentence:

    Returns:

    """
    kw = []
    try:
        temp = requests.get(keywords_url + sentence)
        body = temp.json()
        tag_result = body['SemanticResult']['tagResults']
        for tw in tag_result:
            kw.append(tw['word'])
        labeling = body['KeywordLabeling']['keyword']
        kw.extend(labeling)
        kw = list(set(kw))
    except:
        return []
    return kw


@app.route('/chat', methods=['GET'])
def get_final():
    """
    获取聊天回复
    Returns:

    """
    context = request.args.get('context')
    us = context.strip().split(sep)
    # 获取查询关键词
    kw_q2 = get_keywords(us[-1])
    print("query: ", us[-1])
    print("keywords: ", kw_q2)
    # 语义近似召回
    _, topk, t_vec, t_pca, t_search = indexer.recall(us[-1:], pca, index, k_index)
    topk = list(map(str, topk[0]))
    t0 = time.time()
    # 文本近似召回
    if not kw_q2:
        responses1, origins1 = es.recall(us[-1], topk, k_es)
    else:
        responses1, origins1 = es.recall(' '.join(kw_q2), topk, k_es)
    t01 = time.time()
    responses2, origins2 = es.recall_ids(topk[:10])
    responses = responses1 + responses2
    origins = origins1 + origins2
    t1 = time.time()
    if not responses:
        best, score = 'NA', 0
    else:
        contexts = ['。'.join(us)] * len(responses)
        # 语义匹配排序
        best, score = matcher.predict(contexts, responses)
        print(best)
    t2 = time.time()
    result = json.dumps({'response': best,
                         'score': score,
                         'keywords': kw_q2,
                         'candidates': [o + sep + r for o, r in zip(origins, responses)],
                         'time': t2 - t0 + t_vec + t_pca + t_search,
                         't_vec': t_vec,
                         't_pca': t_pca,
                         't_search': t_search,
                         't_esid': t1 - t01,
                         't_esrecall': t01 - t0,
                         't_input': 0,
                         't_model': 0,
                         't_rank': t2 - t1
                         }, ensure_ascii=False)
    print(result)
    return result


def deploy(config_info):
    global es, sep, index, pca, k_index, k_es, indexer, matcher
    k_index = config_info['recall']['index_topn']
    k_es = config_info['recall']['es_topn']
    sep = "||"

    os.environ["CUDA_VISIBLE_DEVICES"] = '3,4,5,6'

    # parser = argparse.ArgumentParser()
    # 注意local rank这个参数，必须要以这种形式指定，即使代码中不使用。因为 launch 工具默认传递该参数
    # parser.add_argument("--local_rank", type=int, default=0)
    # parser.add_argument("--num_gpu", type=int, default=4)
    # args = parser.parse_args()
    local_rank = 4
    num_gpu = 4

    es = ElasticObj(config_info)

    matcher = ErnieRanker(config_info)
    matcher.load_model()

    indexer = FaissStreamSearcher(local_rank, num_gpu, config_info)
    tt0 = time.time()
    index, pca = indexer.build_ivf()
    tt1 = time.time()
    print(f'build index take {tt1 - tt0} seconds')

    app.run(host='0.0.0.0', port=6997, threaded=True)


# 聊天服务部署
if __name__ == '__main__':
    config_file = "../config/config.json"
    with open(config_file, 'r') as f:
        config_info = json.load(f)
    deploy(config_info)



from torch.utils.data import IterableDataset
import pandas as pd


class Q2StreamSubset(IterableDataset):
    def __init__(self, tsv, local_rank, n_replicas):
        """
        Query2的流式子数据集
        Args:
            tsv: 数据文件
            local_rank: rank index
            n_replicas: 共分为几个数据子集
        """
        super(Q2StreamSubset, self).__init__()
        self.tsv_file = tsv
        self.local_rank = local_rank
        self.n_replicas = n_replicas

    def row_gen(self):
        """
        通过pandas chunksize，流式读取数据
        Returns:
            iterator
        """
        reader = pd.read_csv(self.tsv_file, sep='\t', header=0, na_filter=False, iterator=True, chunksize=1)
        count = -1
        for chunk in reader:
            row = chunk.iloc[0]
            if row['visible'] == 0:
                continue
            count += 1
            if count % self.n_replicas == self.local_rank:
                yield row['q2']

    def __iter__(self):
        return self.row_gen()



"""
独立的多轮聊天服务
"""
import tornado.httpserver
import tornado.ioloop
from tornado.options import define
import tornado.web
import os
import json
import time
import argparse
import random
import requests
from collections import defaultdict
import sys

sys.path.append("/tmp/pycharm_project_201/ptmchat_develop/")
from recall.semantic_search import FaissStreamSearcher
from recall.es_search import ElasticObj
from rank.ranker import ErnieRanker

keywords_url = 'http://10.18.217.184:8077/queryprocv2?userId=yuewenhao&type=TV&query='


def get_keywords(sentence):
    kw = []
    try:
        temp = requests.get(keywords_url + sentence)
        body = temp.json()
        tag_result = body['SemanticResult']['tagResults']
        for tw in tag_result:
            kw.append(tw['word'])
        labeling = body['KeywordLabeling']['keyword']
        kw.extend(labeling)
        kw = list(set(kw))
    except:
        return []
    return kw


# 兜底语句
def getfile(file):
    with open(file, 'r', encoding="utf8") as fr:
        data = fr.readlines()
    return map(lambda x: ''.join(x.strip().split()[1:]), data)


class ChatService:
    def __init__(self, config_info, args):
        self.doudi_response = list(getfile(config_info['inputs']['doudi_response_file']))
        self.es = ElasticObj(config_info)
        self.k_index = config_info['recall']['index_topn']
        self.k_es = config_info['recall']['es_topn']
        self.max_seq_num = 16

        self.matcher = ErnieRanker(config_info)
        self.matcher.load_model()

        self.indexer = FaissStreamSearcher(args, config_info)
        t0 = time.time()
        self.index, self.pca = self.indexer.build_ivf()
        t1 = time.time()
        print(f'build index take {t1 - t0} seconds')

    def chat(self, userID, q2):
        """
        :param userID: 用户id或设备id，用于区分不同用户的聊天历史
        :param q2: 用户当前query str type
        :return:
        """
        global history
        # if userID not in history:
        #     raise KeyError('userID not registered yet!')
        if len(history[userID]) > 1 and history[userID][-2][0] == "再见":
            # 清掉之前的对话历史
            history[userID] = []

        if len(history[userID]) > self.max_seq_num:
            # 顶掉较早的历史
            history[userID].pop(0)
            history[userID].pop(0)

        if len(history[userID]) >= 2:  # 判断是否是第一轮
            q1, r1 = history[userID][-2][0], history[userID][-1]
        else:
            q1, r1 = "nan", "nan"
        kw_q2 = get_keywords(q2)
        # 召回
        _, topk, t_vec, t_pca, t_search = self.indexer.recall([q2], self.pca, self.index, self.k_index)
        topk = list(map(str, topk[0]))
        if not kw_q2:
            responses1, origins1 = self.es.recall(q2, topk, self.k_es)
        else:
            responses1, origins1 = self.es.recall(' '.join(kw_q2), topk, self.k_es)
        responses2, origins2 = self.es.recall_ids(topk[:10])
        responses = responses1 + responses2
        origins = origins1 + origins2
        # rank
        if not responses:
            score = 0
            best = random.choice(self.doudi_response)
        else:
            contexts = ['。'.join([q1, r1, q2])] * len(responses)
            best, score = self.matcher.predict(contexts, responses)
        # 本轮写入历史
        candidates = [o + '||' + r for o, r in zip(origins, responses)]
        history[userID].append((q2, candidates, kw_q2, score))
        history[userID].append(best)
        return history[userID]


class ResultHandler(tornado.web.RequestHandler):
    def get(self):
        last_query = self.get_argument('query')
        userID = self.get_argument("userID")
        history[userID] = chatter.chat(userID, last_query)
        result = {"userID": userID}
        for i in range(len(history[userID])):
            if i % 2:
                # bot
                key = f"chatbot_turn_{i // 2 + 1}"
                result[key] = history[userID][i]
            else:
                # user
                key = f"user_turn_{i // 2 + 1}"
                result[key] = history[userID][i][0]
                if userID == 'test' and i == len(history[userID]) - 2:
                    # 测试账号，显示更多细节
                    result[f"score_turn_{i // 2 + 1}"] = history[userID][i][3]
                    result[f"topn_turn_{i // 2 + 1}"] = history[userID][i][1]
                    result[f"keywords_turn_{i // 2 + 1}"] = history[userID][i][2]
        self.write(json.dumps(result, ensure_ascii=False, indent=1))


if __name__ == '__main__':
    # 导入配置文件
    config_file = "../config/config.json"
    with open(config_file, 'r') as f:
        config_info = json.load(f)

    os.environ["CUDA_VISIBLE_DEVICES"] = '3'

    parser = argparse.ArgumentParser()
    # 注意local rank这个参数，必须要以这种形式指定，即使代码中不使用。因为 launch 工具默认传递该参数
    parser.add_argument("--local_rank", type=int, default=0)
    parser.add_argument("--num_gpu", type=int, default=4)
    args = parser.parse_args()

    # history的key为用户id或设备id，自定义
    history = defaultdict(list)
    chatter = ChatService(config_info, args)

    # web service
    tornado.options.parse_command_line()
    app = tornado.web.Application(
        handlers=[
            (r'/subj_extract', ResultHandler),
        ],  # 网页路径控制
    )
    http_server = tornado.httpserver.HTTPServer(app)
    http_server.listen(6996)
    tornado.ioloop.IOLoop.instance().start()



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
"""
@File        :  setup.py
@Project     :  ptmchat_develop
@Contact     :  yangshansong
@License     :  (C)Copyright 2021-2022, Hisense
@Description :  None

@Modify Time        @Author         @Version        @Description
------------        -------         --------        ------------
2021/12/6 11:30     yangshansong    1.0             None
"""

import os
import setuptools

DIR = os.path.dirname(__file__)
REQUIREMENTS = os.path.join(DIR, "requirements.txt")

with open(REQUIREMENTS) as f:
    reqs = f.read()

setuptools.setup(
    name="ptmchat",
    version="0.0.1",
    description="pretrained ernie model based retrieval open domain chatting",
    url="http://10.18.203.116/dialog/python-aispeech-ptmchat-api",
    author="hisense",
    license="Apache License 2.0",
    packages=setuptools.find_packages(),
    install_requires=reqs.strip().split("\n"),
    entry_points={
        'console_scripts': ['ptmchat=ptmchat_develop.ptmchat_entry:main'],
    }
)



# coding=utf-8
# WebURL：http://10.18.218.206:9100/
from elasticsearch import Elasticsearch
from elasticsearch.helpers import bulk
import time
import pandas as pd
import json


class ElasticObj:
    def __init__(self, config):
        # es settings
        self.index_type = config['es_info']['index_type']
        self.host = config['es_info']['es_host']
        self.user_name = config['es_info']['user_name']
        self.password = config['es_info']['password']
        self.es = Elasticsearch(hosts=[self.host], http_auth=(self.user_name, self.password), timeout=5000)
        self.index_alias = config['es_info']['index_alias']
        if self.es.indices.exists_alias(name=[self.index_alias]):
            self.old_index_name = list(self.es.indices.get_alias(name=[self.index_alias]).keys())[0]
        else:
            self.old_index_name = None
        self.index_name = config['es_info']['index_name_pre'] + time.strftime("%Y%m%d%H%M")

    def set_mapping(self):
        """
        设置映射，创建索引
        Returns:

        """
        body = {
            "settings": {
                "number_of_shards": 3,
                "number_of_replicas": 1
            },
            "mappings": {
                self.index_type: {
                    "properties": {
                        "query1": {"type": "text", "analyzer": "ik_smart", "search_analyzer": "ik_max_word"},
                        "response1": {"type": "text", "analyzer": "ik_smart", "search_analyzer": "ik_max_word"},
                        "query2": {"type": "text", "analyzer": "ik_smart", "search_analyzer": "ik_max_word"},
                        "response2": {"type": "text", "analyzer": "ik_smart", "search_analyzer": "ik_max_word"}
                    }
                }
            }
        }
        self.es.indices.create(index=self.index_name, ignore=True, body=body)

    def gen_file(self, file):
        """
        定义ES动作迭代器，数据插入
        从文件中流式生成每一条数据，并自带ID从0开始
        Args:
            file: tsv原始数据文件

        Returns:

        """
        # 流式读取CSV数据文件
        reader = pd.read_csv(file, sep='\t', header=0, na_filter=False, iterator=True, chunksize=1)
        count = -1  # 文件ID
        for chunk in reader:
            row = chunk.iloc[0]
            if row['visible'] == 0:
                continue
            count += 1
            # 定义数据文档
            document = {
                "query1": row['q1'],
                "response1": row['r1'],
                "query2": row['q2'],
                "response2": row['r2']
            }
            # 定义操作
            action = {
                "_index": self.index_name,
                "_type": self.index_type,
                "_id": count,
                "_source": document
            }
            yield action

    def write_data(self, source):
        """
        将数据写入索引
        Args:
            source:

        Returns:

        """
        # 创建数据操作迭代器
        gen = self.gen_file(source)
        t0 = time.time()
        success, _ = bulk(self.es, gen, index=self.index_name, raise_on_error=True)
        t1 = time.time()
        print(f'performed {success} actions in {(t1 - t0) / 60} minutes.')

    def shift_alias(self):
        """
        Shift the alias from old index to new index
        Returns:

        """
        if self.old_index_name is None:
            body = {
                "actions": [
                    {"add": {"index": self.index_name, "alias": self.index_alias}}
                ]
            }
            # 更新索引别名
            self.es.indices.update_aliases(body=body)
        else:
            body = {
                "actions": [
                    {"remove": {"index": self.old_index_name, "alias": self.index_alias}},
                    {"add": {"index": self.index_name, "alias": self.index_alias}}
                ]
            }
            # 更新索引别名
            self.es.indices.update_aliases(body=body)
            # 删除旧的索引
            self.es.indices.delete(index=self.old_index_name, ignore=[400, 404])

    def index_exist(self):
        """
        索引是否存在
        Returns:

        """
        return self.es.indices.exists(index=[self.index_name])

    @property
    def size(self):
        return self.es.count(index=self.index_name)

    def recall_ids(self, ids):
        """
        按id检索。
        :param ids: List[str] 文档的id列表
        :return: 所有文档
        """
        body = {
            "ids": ids
        }
        sources = ['query1', 'response1', 'query2', 'response2']
        # 索引查询是按照别名来指定具体索引
        hits = self.es.mget(body=body, index=self.index_alias, doc_type=self.index_type, _source=sources)
        all_sources = hits['docs']
        results = []
        contexts = []
        for hit in all_sources:
            if not hit['found']:
                raise KeyError(f"doc id {hit['_id']} not found in ES. Faiss and ES data may not be consistent!")
            result = hit['_source']['response2']
            results.append(result)
            context = hit['_source']['query1'] + '||' + hit['_source']['response1'] + '||' + hit['_source']['query2']
            contexts.append(context)
        return results, contexts

    def recall(self, query, ids, k):
        """
        按id和query检索。
        :param query: 一个query
        :param ids: List[str] 文档的id列表
        :param k: 召回几个
        :return: 所有文档
        """
        body = {
            "query": {
                "bool": {
                    "must": [
                        {
                            "match": {
                                "query2": query
                            }
                        }
                    ],
                    "filter": [
                        {
                            "ids": {
                                "values": ids
                            }
                        }
                    ]
                }
            }
        }
        sources = ['query1', 'response1', 'query2', 'response2']
        # 索引查找，按照别名来指定具体索引
        hits = self.es.search(
            index=self.index_alias,
            body=body,
            _source=sources,
            size=k,
            track_total_hits=False,
            timeout="35ms")
        all_sources = hits['hits']['hits']
        results = []
        contexts = []
        count = len(all_sources)
        for i in range(count):
            source = all_sources[i]['_source']
            result = source['response2']
            results.append(result)
            context = source['query1'] + '||' + source['response1'] + '||' + source['query2']
            contexts.append(context)
        return results, contexts

    def recall_ext(self, keys, extends, ids, k):
        """
        按id和query关键词检索。扩展关键词。
        :param keys: query的关键词
        :param extends: 扩展的关键词 todo 扩展关键词，应该在context中出现
        :param ids: List[str] 文档的id列表
        :param k: 召回几个
        :return: 所有文档
        """
        body = {
            "query": {
                "bool": {
                    "must": [
                        {
                            "match": {
                                "query2": ' '.join(keys + extends)
                            }
                        }
                    ],
                    "should": [
                        {
                            "multi_match": {
                                "query": ' '.join(keys),
                                "fields": ["query1", "response1", "query2^2"]
                            }
                        }
                    ],
                    "filter": [
                        {
                            "ids": {
                                "values": ids
                            }
                        }
                    ]
                }
            }
        }
        sources = ['query1', 'response1', 'query2', 'response2']
        hits = self.es.search(
            index=self.index_alias,
            body=body,
            _source=sources,
            size=k,
            track_total_hits=False,
            timeout="35ms")
        all_sources = hits['hits']['hits']
        results = []
        contexts = []
        count = len(all_sources)
        for i in range(count):
            source = all_sources[i]['_source']
            result = source['response2']
            results.append(result)
            context = source['query1'] + '||' + source['response1'] + '||' + source['query2']
            contexts.append(context)
        return results, contexts

    def test_analyzer(self, query, analyzer='ik_max_word'):
        """
        test the analyzer
        Args:
            query:
            analyzer:

        Returns:

        """
        body = {
            "analyzer": analyzer,
            "text": query
        }
        return self.es.indices.analyze(body=body, index='chat_all202106241410')

    def build_index(self, source):
        """
        构建索引
        Args:
            source:

        Returns:

        """
        # 设置映射，创建索引
        self.set_mapping()
        # 插入数据
        self.write_data(source)
        # 切换别名，删除旧索引
        self.shift_alias()


if __name__ == "__main__":
    with open('../config/config.json', 'r') as f:
        config_info = json.load(f)
    es = ElasticObj(config_info)
    # data = '../data/ernie_clean.tsv'
    # es.build_index(data)
    # res = es.recall_ids(['3205306', '8285888', '10720332', '12040142'])
    res = es.recall('会所那边的情况', ['3205306', '8285888', '10720332', '12040142'], 3)
    print(res)
    # print(es.test_analyzer("会所那边的情况", analyzer='ik_max_word'))
